/******************************************************************************
 * Copyright 2018 NVIDIA Corporation. All rights reserved.
 *****************************************************************************/


#include "mdl_cache_module.h"
#include <mi/mdl_sdk.h>
#include <iostream>
#include "mdl_cache_material.h"
#include "mdl_cache_function.h"
#include "../utilities/platform_helper.h"
#include <set>


bool Mdl_cache_module::update(mi::neuraylib::INeuray* neuray, 
                              mi::neuraylib::ITransaction* transaction, 
                              const mi::base::IInterface* node)
{
    const mi::base::Handle<const mi::neuraylib::IMdl_module_info> module_info(
        node->get_interface<const mi::neuraylib::IMdl_module_info>());
    if (!module_info)
        return false;

    // get the resolved path of the file this module is defined in
    // in case of modules in archives, we need the file path of the archive
    const mi::base::Handle<const mi::IString> resolved_path(module_info->get_resolved_path());
    std::string current_path = resolved_path->get_c_str();
    if(module_info->in_archive())
    {
        const size_t pos = current_path.find(".mdr:");
        current_path = current_path.substr(0, pos + 4);
    }

    // if the search path is one we stored in the cache, we check the date
    const mi::Uint64 timestamp = static_cast<time_t>(
        Platform_helper::get_file_change_time(current_path));

    // maybe, this module is from another search path
    const char* cached_path = get_file_path();
    if(cached_path && strcmp(current_path.c_str(), cached_path) == 0)
    {
        // return true here if the file has not changed
        if (timestamp > 0 && timestamp <= get_timestamp())
        {
            /*
            std::cerr << "[Mdl_cache_module] update: skipped unchanged module: "
                      << get_simple_name() << "\n";
            */
            return true;
        }
    }

    // load the selected module
    mi::base::Handle<mi::neuraylib::IMdl_compiler> compiler(
        neuray->get_api_component<mi::neuraylib::IMdl_compiler>());

    if (!compiler || 0 > compiler->load_module(transaction, module_info->get_qualified_name()))
    {
        std::cerr << "[Mdl_cache_module] update: Failed to load module: " 
                  << get_simple_name() << "\n";
        return false;
    }

    mi::base::Handle<const mi::neuraylib::IModule> mdl_module(
        transaction->access<mi::neuraylib::IModule>(
        (std::string("mdl") + module_info->get_qualified_name()).c_str()));

    if (!mdl_module)
    {
        std::cerr << "[Mdl_cache_module] update: Failed to load module: " 
                  << get_simple_name() << "\n";
        return false;
    }

    // get infos from annotations
    const mi::base::Handle<const mi::neuraylib::IAnnotation_block> anno_block(
        mdl_module->get_annotations());

    // are there annotations?
    if (anno_block)
    {
        const mi::neuraylib::Annotation_wrapper annotations(anno_block.get());

        if (annotations.get_annotation_index("::anno::hidden()") != static_cast<mi::Size>(-1))
            set_is_hidden(true);
    }

    // keep track of the existence of children to allow clean-up
    std::set<Child_map_key> not_present_children;
    for (const auto& c : get_children())
        not_present_children.insert(c.first);

    bool success = true;

    // Iterate over all materials exported by the module.
    for (mi::Size i = 0, n = mdl_module->get_material_count(); i < n; ++i)
    {
        std::string qualified_name = mdl_module->get_material(i) + 3; // ignore the leading "mdl"
        const std::string simple_name = Mdl_helper::qualified_to_simple_name(qualified_name);
        const Child_map_key key{IMdl_cache_item::CK_MATERIAL, simple_name};

        // "mark" as present by removing from not_present_children
        const auto& pos = not_present_children.find(key);
        if (pos != not_present_children.end())
            not_present_children.erase(pos);

        // insert new child if not already available
        IMdl_cache_item* child = get_child(key);
        if (!child)
        {
            child = get_cache()->create(CK_MATERIAL, qualified_name.c_str());
            add_child(child);
        }

        // update recursively
        if (!dynamic_cast<Mdl_cache_material*>(child)->update(
            neuray, transaction, mdl_module.get()))
        {
            std::cerr << "[Mdl_cache_module] update: Failed to update material: " 
                      << get_qualified_name() << "\n";
            success = false;
        }

        // pass down the hidden property
        if (get_is_hidden())
            dynamic_cast<Mdl_cache_material*>(child)->set_is_hidden(true);
    }

    // Iterate over all materials exported by the module.
    for (mi::Size i = 0, n = mdl_module->get_function_count(); i < n; ++i)
    {
        std::string qualified_name = mdl_module->get_function(i) + 3; // ignore the leading "mdl"
        const std::string simple_name = Mdl_helper::qualified_to_simple_name(qualified_name);
        const Child_map_key key{IMdl_cache_item::CK_FUNCTION, simple_name};

        // "mark" as present by removing from not_present_children
        const auto& pos = not_present_children.find(key);
        if (pos != not_present_children.end())
            not_present_children.erase(pos);

        // insert new child if not already available
        IMdl_cache_item* child = get_child(key);
        if (!child)
        {
            child = get_cache()->create(CK_FUNCTION, qualified_name.c_str());
            add_child(child);
        }

        // update recursively
        if (!dynamic_cast<Mdl_cache_function*>(child)->update(
            neuray, transaction, mdl_module.get()))
        {
            std::cerr << "[Mdl_cache_module] update: Failed to update function: " 
                      << get_qualified_name() << "\n";
            success = false;
        }

        // pass down the hidden property
        if (get_is_hidden())
            dynamic_cast<Mdl_cache_function*>(child)->set_is_hidden(true);
    }

    // remove all cached modules that are not present anymore
    for (const auto& c : not_present_children)
    {
        IMdl_cache_item* item = remove_child(c);
        get_cache()->erase(item);
    }
        
    // keep the timestamp if everything went fine
    set_timestamp(success ? timestamp : 0);
    set_file_path(current_path.c_str()); // store the search path, too
    set_located_in_archive(module_info->in_archive());

    return success;
}

bool Mdl_cache_module::get_located_in_archive() const
{
    const char* value = Base::get_cache_data("LocatedInArchive");
    return value && strcmp(value, "true") == 0;
}

void Mdl_cache_module::set_located_in_archive(bool value)
{
    Base::set_cache_data("LocatedInArchive", value ? "true" : "false");
}

const char* Mdl_cache_module::get_file_path() const
{
    return Base::get_cache_data("FilePath");
}

void Mdl_cache_module::set_file_path(const char* search_path)
{
    Base::set_cache_data("FilePath", search_path);
}
